
import numpy as np

import matplotlib.pyplot as plt 
import matplotlib.cm as cm
from matplotlib.colors import Normalize
from matplotlib.ticker import NullLocator


def plot_pop_scores(scores, title = 'Population scores', save_as = None):
    """
    plots scores (numeric values) as sorted bars
    """
    
    sc = sorted(scores)[::-1]

    fig, ax = plt.subplots(figsize=(7, 5))
    ax.xaxis.set_major_locator(NullLocator())

    def autolabel(rects):
        """Attach a text label above each bar in *rects*, displaying its height."""
        for rect in rects[-1:]:
            height = round(rect.get_height(), 2)
            ax.annotate('{}'.format(height),
                        xy = (rect.get_x() + rect.get_width() / 2, height),
                        xytext=(0, 3),  # 3 points vertical offset
                        textcoords="offset points",
                        ha='center', va='bottom', fontsize=14, fontweight='bold')
    


    cols = np.zeros(len(sc))
    cols[-1] = 1

    x_coord = np.arange(len(sc))
    my_norm = Normalize(vmin=0, vmax=1)
    
    rc = ax.bar(x_coord, sc,  width = 0.7, color = cm.get_cmap('Set2')(my_norm(cols)))
    
    autolabel(rc)

    #ax.set_xticks(x_coord)
    ax.set_title(title, fontsize=16, fontweight='bold')
    ax.set_xlabel('Population objects')
    ax.set_ylabel('Cost values')
    #ax.set_ylim([0, max(subdict.values())*1.2])
    #fig.suptitle(title, fontsize=15, fontweight='bold')

    
    fig.tight_layout()
    
    if not (save_as is None):
        plt.savefig(save_as, dpi = 200)
    
    plt.show()




import math
import random
import numpy as np



class Selection:
    
    def __inverse_scores(scores):
        """
        inverse scores (min val goes to max)
        """
        minobj = scores[0]
        normobj = scores - minobj if minobj < 0 else scores
                
        return (np.amax(normobj) + 1) - normobj 
    
    
    def fully_random():
        
        def func(scores, parents_count):
            indexes = np.arange(parents_count)
            return np.random.choice(indexes, parents_count, replace = False)
        
        return func
    
    def __roulette(scores, parents_count):
        
        sum_normobj = np.sum(scores)
        prob = scores/sum_normobj
        cumprob = np.cumsum(prob)            
            
        parents_indexes = np.empty(parents_count)
            
        # it can be vectorized
        for k in range(parents_count):
            index = np.searchsorted(cumprob, np.random.random())
            if index < cumprob.size:
                parents_indexes[k] = index
            else:
                parents_indexes[k] = np.random.randint(0, index - 1)
            
        return parents_indexes

    
    def roulette():
        
        def func(scores, parents_count):

            normobj = Selection.__inverse_scores(scores)

            return Selection.__roulette(normobj, parents_count)
        
        return func
    
    def stochastic():
        
        def func(scores, parents_count):
            f = Selection.__inverse_scores(scores)
            
            fN = 1.0/parents_count
            k = 0
            acc = 0.0
            parents = []
            r = random.random()*fN
            
            while len(parents) < parents_count:
                
                acc += f[k]
                
                while acc > r:
                    parents.append(k)
                    if len(parents) == parents_count: break
                    r += fN
                
                k += 1
            
            return np.array(parents[:parents_count])
        
        return func
    
    def sigma_scaling(epsilon = 0.01, is_noisy = False):
        
        def func(scores, parents_count):
            f = Selection.__inverse_scores(scores)
            
            sigma = np.std(f, ddof = 1) if is_noisy else np.std(f)
            average = np.mean(f)
            
            if sigma == 0:
                f = 1
            else:
                f = np.maximum(epsilon, 1 + (f - average)/(2*sigma))
            
            return Selection.__roulette(f, parents_count)
        
        return func
    
    def ranking():
        
        def func(scores, parents_count):
            return Selection.__roulette(1 + np.arange(parents_count)[::-1], parents_count)
        
        return func
    
    def linear_ranking(selection_pressure = 1.5):
        
        assert (selection_pressure > 1 and selection_pressure < 2), f"selection_pressure should be in (1, 2), but got {selection_pressure}"
        
        def func(scores, parents_count):
            tmp = parents_count*(parents_count-1)
            alpha = (2*parents_count - selection_pressure*(parents_count + 1))/tmp
            beta = 2*(selection_pressure - 1)/tmp
            
            
            a = -2*alpha - beta
            b = (2*alpha + beta)**2
            c = 8*beta
            d = 2*beta
            
            indexes = np.arange(parents_count)
            
            return np.array([indexes[-round((a + math.sqrt(b + c*random.random()))/d)] for _ in range(parents_count)])
            
        
        return func
    
    def tournament(tau = 2):
        
        def func(scores, parents_count):
            
            indexes = np.arange(parents_count)
            
            return np.array([np.min(np.random.choice(indexes, tau, replace = False)) for _ in range(parents_count)])
            
        
        return func
    
    
    





















import numpy as np

from functools import lru_cache, wraps
#from fastcache import clru_cache



def np_lru_cache(*args, **kwargs):
    """
    LRU cache implementation for functions whose FIRST parameter is a numpy array
        forked from: https://gist.github.com/Susensio/61f4fee01150caaac1e10fc5f005eb75
    """

    def decorator(function):
        
        @wraps(function)
        def wrapper(np_array):
            return cached_wrapper( tuple(np_array))

        @lru_cache(*args, **kwargs)
        #@clru_cache(*args, **kwargs)
        def cached_wrapper(hashable_array):
            return function(np.array(hashable_array))

        # copy lru_cache attributes over too
        wrapper.cache_info = cached_wrapper.cache_info
        wrapper.cache_clear = cached_wrapper.cache_clear
        return wrapper
    
    return decorator





if __name__ == '__main__':

    ar = np.random.randint(0, 100, (1000, 100))

    f = lambda arr: np.std(arr + arr/(1 + arr**2) - arr + np.sin(arr) * np.cos(arr) + 2)

    def no_c(arr):
        return f(arr)
    
    @np_lru_cache(maxsize = 700, typed = True)
    def with_c(arr):
        return f(arr)

    #%time for _ in range(50): [no_c(arr) for arr in ar[np.random.rand(ar.shape[0]).argsort()]]
    #%time for _ in range(50): [with_c(arr) for arr in ar[np.random.rand(ar.shape[0]).argsort()]]








import math
import random
import numpy as np

class Mutations:
    
    def uniform_by_x():
        
        def func(x, left, right):
            alp = min(x - left, right - x)
            return np.random.uniform(x - alp, x + alp)
        return func
    
    
    def uniform_by_center():
        
        def func(x, left, right):
            return np.random.uniform(left, right)
        
        return func
    
    def gauss_by_x(sd = 0.3):
        """
        gauss mutation with x as center and sd*length_of_zone as std
        """
        def func(x, left, right):
            std = sd * (right - left)
            return max(left, min(right, np.random.normal(loc = x, scale = std)))
        
        return func
    
    def gauss_by_center(sd = 0.3):
        """
        gauss mutation with (left+right)/2 as center and sd*length_of_zone as std
        """
        def func(x, left, right):
            std = sd * (right - left)
            return max(left, min(right, np.random.normal(loc = (left+right)*0.5, scale = std)))
        
        return func



import os
import random

import numpy as np 

import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from OppOpPopInit import OppositionOperators, SampleInitializers


def folder_create(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)


class Callbacks:
    
    @staticmethod
    def NoneCallback():
        return lambda generation_number, report_list, last_population, last_scores: None


    @staticmethod
    def SavePopulation(folder, save_gen_step = 50, file_prefix = 'population'):
        
        folder_create(folder)

        def func(generation_number, report_list, last_population, last_scores):
    
            if generation_number % save_gen_step != 0:
                return

            np.savez(os.path.join(folder, f"{file_prefix}_{generation_number}.npz"), population=last_population, scores=last_scores)
        
        return func
    
    @staticmethod
    def PlotOptimizationProcess(folder, save_gen_step = 50, show = False, main_color = 'green', file_prefix = 'report'):
        folder_create(folder)

        def func(generation_number, report_list, last_population, last_scores):

            if generation_number % save_gen_step != 0:
                return

            # if len(report_list) == 0:
            #     sys.stdout.write("No results to plot!\n")
            #     return
            
            ax = plt.axes()
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            
            plt.plot(np.arange(1, 1 + len(report_list)), report_list, color = main_color, label = 'best of generation', linewidth = 2)
            
            plt.xlabel('Generation')
            plt.ylabel('Minimized function')
            plt.title('GA optimization process')
            plt.legend()
            
            plt.savefig(os.path.join(folder, f"{file_prefix}_{generation_number}.png"), dpi = 200)

            if show: plt.show()
            else: plt.close()
        
        return func






class Actions:

    @staticmethod
    def Stop():
        
        def func(data):
            data['current_stagnation'] = 2*data['max_stagnation']
            return data
        return func

    @staticmethod
    def ReduceMutationProb(reduce_coef = 0.9):
        
        def func(data):
            data['mutation_prob'] *= reduce_coef
            return data
        
        return func


    #def DualStrategyStep():
    #    pass

    #def SetFunction():
    #    pass



    @staticmethod
    def ChangeRandomCrossover(available_crossovers):

        def func(data):
            data['crossover'] = random.choice(available_crossovers)
            return data

        return func
    
    @staticmethod
    def ChangeRandomSelection(available_selections):

        def func(data):
            data['selection'] = random.choice(available_selections)
            return data

        return func

    @staticmethod
    def ChangeRandomMutation(available_mutations):

        def func(data):
            data['mutation'] = random.choice(available_mutations)
            return data

        return func
    


    @staticmethod
    def RemoveDuplicates(oppositor = None, creator = None, converter = None):
        """
        Removes duplicates from population

        Parameters
        ----------
        oppositor : oppositor from OppOpPopInit, optional
            oppositor for applying after duplicates removing. By default -- using just random initializer from creator. The default is None.
        creator : the function creates population samples, optional
            the function creates population samples if oppositor is None. The default is None.
        converter : func, optional
            function converts population samples in new format to compare (if needed). The default is None.

        Raises
        ------
        Exception
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """

        if creator is None and oppositor is None:
            raise Exception("No functions to fill population! creator or oppositors must be not None")

        if converter is None:
            def without_dup(pop, scores): # returns population without dups
                _, index_of_dups = np.unique(pop, axis = 0, return_index = True) 
                return np.hstack((pop[index_of_dups,:], scores[index_of_dups].reshape(-1, 1))), pop.shape[0] - index_of_dups.size
        else:
             def without_dup(pop, scores): # returns population without dups
                _, index_of_dups = np.unique(np.array([converter(pop[i]) for i in range(pop.shape[0])]), axis = 0, return_index = True) 
                return np.hstack((pop[index_of_dups,:], scores[index_of_dups].reshape(-1, 1))), pop.shape[0] - index_of_dups.size           


        if oppositor is None:
            def remover(pop, scores, set_function):

                pp, count_to_create = without_dup(pop, scores) # pop without dups
                pp2 = np.empty((count_to_create, pp.shape[1])) 
                pp2[:,:-1] = SampleInitializers.CreateSamples(creator, count_to_create) # new pop elements
                pp2[:, -1] = set_function(pp2[:,:-1]) # new elements values
                    
                new_pop = np.vstack((pp, pp2))

                return new_pop[np.argsort(new_pop[:,-1]),:] # new pop

        else: # using oppositors
            def remover(pop, scores, set_function):

                pp, count_to_create = without_dup(pop, scores) # pop without dups

                if count_to_create > pp.shape[0]:
                    raise Exception("Too many duplicates, cannot oppose")
                
                if count_to_create == 0:
                    return pp[np.argsort(pp[:, -1]),:]
                
                pp2 = np.empty((count_to_create, pp.shape[1])) 
                # oppose count_to_create worse elements
                pp2[:,:-1] = OppositionOperators.Reflect(pp[-count_to_create:,:-1], oppositor)# new pop elements
                pp2[:, -1] = set_function(pp2[:,:-1]) # new elements values
                    
                new_pop = np.vstack((pp, pp2))
                    
                return new_pop[np.argsort(new_pop[:,-1]),:] # new pop


        def func(data):
            new_pop = remover(data['last_generation']['variables'], data['last_generation']['scores'], data['set_function'])
            
            data['last_generation']['variables'] = new_pop[:,:-1]
            data['last_generation']['scores'] = new_pop[:,-1]

            return data

        
        return func



class ActionConditions:
    
    @staticmethod
    def EachGen(generation_step = 10):
        def func(data):
            return data['current_generation'] % generation_step == 0 and data['current_generation'] > 0
        return func

    @staticmethod
    def AfterStagnation(stagnation_generations = 50):

        def func(data):
            return data['current_stagnation'] % stagnation_generations == 0 and data['current_stagnation'] > 0
        return func


    @staticmethod
    def Several(list_of_conditions):
        """
        returns function which checks all conditions from list_of_conditions
        """

        def func(data):
            return all((cond(data) for cond in list_of_conditions))
        
        return func




class MiddleCallbacks:

    @staticmethod
    def UniversalCallback(action, condition):
        
        def func(data):

            cond = condition(data)
            if cond:
                data = action(data)

            return data, cond
        
        return func

    #def ReduceMutationGen(reduce = 0.9, each_generation = 50):
    #    pass

    #def ReduceMutationStagnation(reduce = 0.5, stagnation_gens = 50):
    #    pass 
       








